/*
Copyright 2018 Gravitational, Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

package docker

import (
	"bytes"
	"crypto/tls"
	"crypto/x509"
	"encoding/pem"
	"fmt"
	"io"
	"io/ioutil"
	"net"
	"net/http"
	"os"
	"path/filepath"
	"strings"
	"time"

	"github.com/gravitational/gravity/lib/defaults"
	"github.com/gravitational/gravity/lib/loc"

	"github.com/docker/distribution"
	"github.com/docker/distribution/context"
	"github.com/docker/distribution/registry/api/errcode"
	registryclient "github.com/docker/distribution/registry/client"
	registrystorage "github.com/docker/distribution/registry/storage"
	"github.com/docker/distribution/registry/storage/cache/memory"
	"github.com/docker/distribution/registry/storage/driver/filesystem"
	"github.com/docker/libtrust"
	"github.com/gravitational/trace"
	log "github.com/sirupsen/logrus"
)

// RegistryConnectionRequest represents connection information for a Docker registry
type RegistryConnectionRequest struct {
	// RegistryAddress is host:port of a Docker registry
	RegistryAddress string
	// CertName allows to override directory where to look for certs
	// which normally equals to the registry address
	CertName string
	// CACertPath is the full path to the root certificate
	CACertPath string
	// ClientCertPath is the full path to the client certificate
	ClientCertPath string
	// ClientKeyPath is the full path to the client private key
	ClientKeyPath string
}

// CheckAndSetDefaults makes sure the request is valid and sets some defaults
func (r *RegistryConnectionRequest) CheckAndSetDefaults() error {
	if r.RegistryAddress == "" {
		return trace.BadParameter("missing RegistryAddress")
	}
	certName := r.CertName
	if certName == "" {
		certName = r.RegistryAddress
	}
	if r.CACertPath == "" {
		r.CACertPath = filepath.Join(
			defaults.DockerCertsDir, certName, certName+".crt")
	}
	if r.ClientCertPath == "" {
		r.ClientCertPath = filepath.Join(
			defaults.DockerCertsDir, certName, "client.cert")
	}
	if r.ClientKeyPath == "" {
		r.ClientKeyPath = filepath.Join(
			defaults.DockerCertsDir, certName, "client.key")
	}
	return nil
}

// String returns string representation for the request
func (r RegistryConnectionRequest) String() string {
	return fmt.Sprintf("Registry(Address=%v, CA=%v, ClientCert=%v, ClientKey=%v)",
		r.RegistryAddress, r.CACertPath, r.ClientCertPath, r.ClientKeyPath)
}

// NewImageService creates an image service using the supplied
// address and certificate name to connect to the remote registry
func NewImageService(req RegistryConnectionRequest) (ImageService, error) {
	err := req.CheckAndSetDefaults()
	if err != nil {
		return nil, trace.Wrap(err)
	}
	return &imageService{
		RegistryConnectionRequest: req,
		FieldLogger:               log.WithField("registry", req.RegistryAddress),
	}, nil
}

// imageService implements ImageService using provided remote registry address
type imageService struct {
	ImageService
	RegistryConnectionRequest
	log.FieldLogger

	remoteStore *remoteStore
}

// Sync synchronizes the contents of the local directory specified with dir
// with the contents of the remote registry.
// dir is expected to be in docker registry 2.x format.
//
// Upon success, returns a list of images pushed to the registry.
func (r *imageService) Sync(ctx context.Context, dir string) (installedTags []TagSpec, err error) {
	if err = r.connect(ctx); err != nil {
		return nil, trace.Wrap(err)
	}
	r.Debugf("Synchronizing local directory %q.", dir)
	localStore, err := openLocal(dir)
	if err != nil {
		return nil, trace.Wrap(err, "failed to open local directory %q as local registry", dir)
	}
	repos, err := ListRepos(ctx, localStore)
	if err != nil {
		return nil, trace.Wrap(err, "failed to list local repositories in %q", dir)
	}

	for _, localRepoName := range repos {
		localRepo, err := localStore.Repository(ctx, localRepoName)
		if err != nil {
			return nil, trace.Wrap(err)
		}
		remoteRepo, err := r.remoteStore.Repository(ctx, localRepoName)
		if err != nil {
			return nil, trace.Wrap(err)
		}
		localManifests, err := localRepo.Manifests(ctx)
		if err != nil {
			return nil, trace.Wrap(err)
		}
		remoteManifests, err := remoteRepo.Manifests(ctx)
		if err != nil {
			return nil, trace.Wrap(err)
		}
		// for every repository enumerate available tags
		localTags := localRepo.Tags(ctx)
		remoteTags := remoteRepo.Tags(ctx)
		// enumerate local repositories
		tags, err := localTags.All(ctx)
		if err != nil {
			return nil, trace.Wrap(err)
		}
		for _, tag := range tags {
			desc, err := localTags.Get(ctx, tag)
			if err != nil {
				return nil, trace.Wrap(err)
			}
			localManifest, err := localManifests.Get(ctx, desc.Digest)
			if err != nil {
				return nil, trace.Wrap(err)
			}
			// see if the remote registry has this reference
			remoteDesc, err := remoteTags.Get(ctx, tag)
			var remoteManifest distribution.Manifest
			if err == nil {
				remoteManifest, err = remoteManifests.Get(ctx, remoteDesc.Digest)
				if err != nil && !IsManifestUnknown(err) {
					return nil, trace.Wrap(err)
				}
			}

			tagSpec := TagSpec{
				Name:    localRepoName,
				Version: tag,
			}
			// remote registry either does not have this reference, or it is
			// different from the local one
			if remoteManifest == nil || !compareManifests(localManifest, remoteManifest) {
				if err = r.remoteStore.updateRepo(ctx, remoteRepo, localRepo, localManifest, tag); err != nil {
					return nil, trace.Wrap(err, "failed to update remote for tag %q", tagSpec)
				}
			}
			installedTags = append(installedTags, tagSpec)
		}
	}
	return installedTags, nil
}

// Wrap translates the specified image to point to the private registry
// this image service is managing if the image is not already pointing to it.
func (r *imageService) Wrap(image string) string {
	parsed, err := loc.ParseDockerImage(image)
	if err != nil {
		return image
	}
	parsed.Registry = r.RegistryAddress
	return parsed.String()
}

// Unwrap translates the specified image to point to the original repository
// Its function is the inverse of Wrap.
func (r *imageService) Unwrap(image string) (unwrapped string) {
	unwrapped = TagFromString(image).String()
	return strings.TrimPrefix(unwrapped, fmt.Sprintf("%v/", r.RegistryAddress))
}

func (r *imageService) connect(ctx context.Context) (err error) {
	if r.remoteStore == nil {
		r.remoteStore, err = ConnectRegistry(ctx, r.RegistryConnectionRequest)
		if err != nil {
			return trace.Wrap(err, "failed to connect to registry at %q", r.RegistryAddress)
		}
	}
	return nil
}

// remoteStore defines a remote distribution registry
type remoteStore struct {
	log.FieldLogger
	transport *http.Transport
	registry  registryclient.Registry
	addr      string
}

// localStore defines a distribution registry from a local directory
type localStore struct {
	registry distribution.Namespace
	dir      string
}

// newCertPool creates x509 certPool with provided CA files.
func newCertPool(CAFiles []string) (*x509.CertPool, error) {
	certPool := x509.NewCertPool()

	for _, CAFile := range CAFiles {
		pemByte, err := ioutil.ReadFile(CAFile)
		if err != nil {
			return nil, trace.Wrap(err)
		}

		for {
			var block *pem.Block
			block, pemByte = pem.Decode(pemByte)
			if block == nil {
				break
			}
			cert, err := x509.ParseCertificate(block.Bytes)
			if err != nil {
				return nil, trace.Wrap(err)
			}
			certPool.AddCert(cert)
		}
	}

	return certPool, nil
}

func initTransport(req RegistryConnectionRequest) (*http.Transport, string) {
	const connectTimeout = 30 * time.Second
	const keepAlivePeriod = 30 * time.Second
	const handshakeTimeout = 30 * time.Second

	transport := &http.Transport{
		Proxy: http.ProxyFromEnvironment,
		Dial: (&net.Dialer{
			Timeout:   connectTimeout,
			KeepAlive: keepAlivePeriod,
			DualStack: true,
		}).Dial,
		TLSHandshakeTimeout: handshakeTimeout,
		DisableKeepAlives:   true,
	}

	cert, err := tls.LoadX509KeyPair(req.ClientCertPath, req.ClientKeyPath)
	if err == nil {
		roots, err := newCertPool([]string{req.CACertPath})
		if err == nil {
			log.Debugf("Found TLS trust for %s.", req)
			transport.TLSClientConfig = &tls.Config{
				Certificates: []tls.Certificate{cert},
				MinVersion:   tls.VersionTLS10,
				RootCAs:      roots,
			}
			return transport, fmt.Sprintf("https://%v", req.RegistryAddress)
		}
	}

	log.Debugf("No TLS trust for %s.", req)
	transport.TLSClientConfig = &tls.Config{InsecureSkipVerify: true}
	return transport, fmt.Sprintf("http://%v", req.RegistryAddress)
}

// ConnectRegistry connects to the registry with the specified address
func ConnectRegistry(ctx context.Context, request RegistryConnectionRequest) (*remoteStore, error) {
	transport, registryAddr := initTransport(request)
	if err := ping(transport, registryAddr); err != nil {
		return nil, trace.Wrap(err)
	}

	registry, err := registryclient.NewRegistry(ctx, registryAddr, transport)
	if err != nil {
		return nil, trace.Wrap(err)
	}

	return &remoteStore{
		FieldLogger: log.WithField("remote registry", registryAddr),
		addr:        registryAddr,
		transport:   transport,
		registry:    registry,
	}, nil
}

func ping(transport *http.Transport, registryAddr string) error {
	const pingClientTimeout = 30 * time.Second
	pingClient := &http.Client{
		Transport: transport,
		Timeout:   pingClientTimeout,
	}
	endpoint := registryAddr + "/v2/"
	req, err := http.NewRequest("GET", endpoint, nil)
	if err != nil {
		return trace.Wrap(err)
	}
	resp, err := pingClient.Do(req)
	if err != nil {
		return trace.Wrap(err)
	}
	defer resp.Body.Close()
	data, err := ioutil.ReadAll(resp.Body)
	if err != nil {
		return trace.Wrap(err)
	}
	log.WithField("registry", registryAddr).Debugf("Repository response: %s.", data)
	return nil
}

// openLocal creates a distribution registry in the local directory given with dir
func openLocal(dir string) (store *localStore, err error) {
	fi, err := os.Stat(dir)
	if err != nil {
		return nil, trace.Wrap(err)
	}
	if !fi.IsDir() {
		return nil, trace.BadParameter("%v not a valid directory", dir)
	}

	key, err := libtrust.GenerateECP256PrivateKey()
	if err != nil {
		return nil, trace.Wrap(err, "failed to generate a signing key")
	}

	driver := filesystem.New(filesystem.DriverParameters{
		RootDirectory: dir,
		MaxThreads:    defaults.ImageServiceMaxThreads,
	})
	cacheProvider := memory.NewInMemoryBlobDescriptorCacheProvider()
	options := []registrystorage.RegistryOption{
		registrystorage.BlobDescriptorCacheProvider(cacheProvider),
		registrystorage.Schema1SigningKey(key),
		registrystorage.EnableDelete,
	}
	ns, err := registrystorage.NewRegistry(context.Background(), driver, options...)
	if err != nil {
		return nil, trace.Wrap(err)
	}
	return &localStore{
		registry: ns,
		dir:      dir,
	}, nil
}

// Repositories lists the remote repositories
func (s *remoteStore) Repositories(ctx context.Context, entries []string, last string) (n int, err error) {
	return s.registry.Repositories(ctx, entries, last)
}

// Repository provides access to the repository named with name
func (s *remoteStore) Repository(ctx context.Context, name string) (distribution.Repository, error) {
	named, err := parseNamed(name)
	if err != nil {
		return nil, trace.Wrap(err, "invalid named reference %q", name)
	}
	return registryclient.NewRepository(ctx, named, s.addr, s.transport)
}

// Repositories lists the local repositories
func (l *localStore) Repositories(ctx context.Context, entries []string, last string) (n int, err error) {
	return l.registry.Repositories(ctx, entries, last)
}

// Repository provides access to the repository named with name
func (l *localStore) Repository(ctx context.Context, name string) (distribution.Repository, error) {
	named, err := parseNamed(name)
	if err != nil {
		return nil, trace.Wrap(err, "invalid named reference %q", name)
	}
	return l.registry.Repository(ctx, named)
}

func ListRepos(ctx context.Context, namespace registryclient.Registry) (repos []string, err error) {
	const pageSize = 50
	var last string
	p := make([]string, pageSize)
	for n := pageSize; n == pageSize && err == nil; {
		n, err = namespace.Repositories(ctx, p, last)
		if (err == nil || err == io.EOF) && n > 0 {
			repos = append(repos, p[:n]...)
			last = repos[len(repos)-1]
		}
	}
	if err == io.EOF && len(repos) > 0 {
		err = nil
	}
	return repos, err
}

// IsManifestUnknown determines if the specified error is an `unknown manifest` error
func IsManifestUnknown(err error) bool {
	return ("MANIFEST_UNKNOWN" == registryErrorCode(err))
}

// registryErrorCode takes an error returned by registry client and tries
// to recover the docker Registry API error code
func registryErrorCode(err error) string {
	if wrapper, ok := err.(errcode.Errors); ok {
		if wrapper.Len() > 0 {
			if innerError, ok := wrapper[0].(errcode.Error); ok {
				return innerError.ErrorCode().String()
			}
		}
	}
	return ""
}

// compareManifests determines if two Docker tag manifests are equal
func compareManifests(a, b distribution.Manifest) bool {
	if a == nil || b == nil {
		return false
	}
	mediaTypeA, payloadA, _ := a.Payload()
	mediaTypeB, payloadB, _ := b.Payload()
	if mediaTypeA != mediaTypeB {
		return false
	}
	return bytes.Equal(payloadA, payloadB)
}

// updateRepo takes a pair of local+remote repositories and makes the remote repo identical
// to the local one.
func (s *remoteStore) updateRepo(ctx context.Context, remote, local distribution.Repository, manifest distribution.Manifest, tag string) error {
	s.Debugf("Pushing %[1]v --> %[2]v/%[1]v.", local.Named(), s.addr)
	remoteManifests, err := remote.Manifests(ctx)
	if err != nil {
		return trace.Wrap(err)
	}
	localBlobs := local.Blobs(ctx)
	remoteBlobs := remote.Blobs(ctx)
	// copy layers:
	for _, localDesc := range manifest.References() {
		desc, err := remoteBlobs.Stat(ctx, localDesc.Digest)
		if err == nil && desc.Digest == localDesc.Digest {
			s.Debugf("Skipping layer %v.", localDesc.Digest)
			continue
		}
		reader, err := localBlobs.Open(ctx, localDesc.Digest)
		if err != nil {
			return trace.Wrap(err)
		}
		defer reader.Close()
		writer, err := remoteBlobs.Create(ctx)
		if err != nil {
			return trace.Wrap(err)
		}
		defer writer.Close()
		s.Debugf("Writing layer %v.", localDesc.Digest)
		written, err := io.Copy(writer, reader)
		if err != nil {
			return trace.Wrap(err)
		}
		_, err = writer.Commit(ctx, distribution.Descriptor{Digest: localDesc.Digest})
		if err != nil {
			return trace.Wrap(err)
		}
		s.Debugf("Written %v bytes.", written)
	}
	s.Debugf("Updating manifest for %v.", local.Named())
	_, err = remoteManifests.Put(ctx, manifest, distribution.WithTag(tag))
	return trace.Wrap(err)
}
